#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import subprocess
import sys
import xml.etree.ElementTree
from collections import defaultdict
from configobj import ConfigObj
from systemd import journal
import re
import time

if 'eventlet.os' in sys.modules:
    lteStream = sys.stdout
else:
    lteStream = journal.stream(identifier='if_lte')

apn_dbase = '/scripts/apn/apns-conf.xml'
apn_mcc_file = '/scripts/apn/apncountrycode.txt'
apn_default = "internet"
user_default = ""
pass_default = ""
modem_default = '/dev/cdc-wdm0'
mbim_modem_default= '/dev/cdc-wdm_pcie'
modem_tty_default = '/dev/ttyUSB2'


QMI_DRV = 1
MHI_DRV = 2
CDC_DRV = 3
MBM_DRV = 4



def get_mcc_codes(dev):
    mcc = "0"
    mnc = "0"
    try:
        output = subprocess.check_output(['/usr/bin/qmicli', '-d', dev, '--nas-get-home-network'])
    except:
        return int(mcc), int(mnc)

    for ln in output.splitlines():
        if bytes('MCC', encoding='utf8') in ln:
            fields = ln.split(b"'")
            mcc = fields[1]
        if bytes('MNC', encoding='utf8') in ln:
            fields = ln.split(b"'")
            mnc = fields[1]
        if bytes('Description', encoding='utf8') in ln:
            fields = ln.split(b"'")

    return int(mcc), int(mnc)


def get_lte_apn(apn_file, mcc_file, mcc, mnc):
    codes = defaultdict(dict)

    with open(mcc_file, encoding="ISO-8859-1") as f:
        for x in f:
            code, country = x.rstrip().split(None, 1)
            codes.update({int(code): country})

    apns = xml.etree.ElementTree.parse(apn_file).getroot()
    if 0 != mcc:
        for acr in apns.findall('apn'):
            if int(acr.get('mcc')) == mcc and int(acr.get('mnc')) == mnc:
                if "default" in acr.get('type'):
                    apn = acr.get('apn')
                    user = acr.get('user')
                    password = acr.get('password')
                    if not apn:
                        apn = apn_default

                    if not user:
                        user = user_default

                    if not password:
                        password = pass_default

                    return apn, user, password

        print("Local operator %d,%d in '%s' not in our database,going to use default apn '%s'" %
              mcc, mnc, codes.get(mcc), apn_default, file=lteStream)
    else:
        print("Failed to get mcc code,going to use default apn '%s'" % apn_default, file=lteStream)

    return apn_default, user_default, pass_default


def get_qmi_authenticators(if_cfg, modem):
    if 'apn' not in if_cfg or not if_cfg['apn']:
        mcc, mnc = get_mcc_codes(modem)
        apn, user, password = get_lte_apn(apn_dbase, apn_mcc_file, mcc, mnc)
    else:
        apn = if_cfg['apn'] if 'apn' in if_cfg else ""
        user = if_cfg['user'] if 'user' in if_cfg else ""
        password = if_cfg['pass'] if 'pass' in if_cfg else ""

    return apn, user, password


def lte_is_qmi(modem_drv):
    qmi_missing = os.system('ls %s 2>/dev/null >/dev/null' % modem_drv)
    if qmi_missing!=0:
        return False
    
    if lte_get_driver_type_by_dev(modem_drv) != QMI_DRV:
        return
        
    qmi_noop=os.system("/usr/bin/qmicli -d %s --device-open-qmi --dms-noop" % modem_drv)
    if qmi_noop==0:
        return True
    return False
    
def lte_is_mbim(modem_drv):
    modem_drv = mbim_dev_check(modem_drv)
    drv_missing = os.system('ls %s 2>/dev/null >/dev/null' % modem_drv)
    if drv_missing!=0:
        return False
    mbim_noop=os.system("/usr/bin/mbimcli -d %s -p  --no-close --noop" % modem_drv)
    if 0==mbim_noop:
        return True
    return False


def lte_qmi_sim_ready(modem_drv):
    retcode = False
    out = ""
    try:
        out = subprocess.check_output(['/usr/bin/qmicli', '-d', modem_drv, '--uim-get-card-status']).decode()
    except:
        pass
    if 'ready' in out.lower():
        retcode = True
    return retcode
    
    
def lte_qmi_is_connected(modem_drv):
    retcode = False
    out = ""
    try:
        out = subprocess.check_output(['/usr/bin/qmicli', '-d', modem_drv, '--wds-get-packet-service-status']).decode()
    except:
        pass
    m = re.match(".*Connection status:.*'(.*)'.*", out)
    if not m:
        #we got timeout or other message , try with sync option
        try:
            out = subprocess.check_output(['/usr/bin/qmicli', '-d', modem_drv, '--wds-get-packet-service-status', '--device-open-sync']).decode()
        except:
            pass
        m = re.match(".*Connection status:.*'(.*)'.*", out)
    if m and m.group(1):
        connst = m.group(1)
        if connst.lower() == "connected":
            retcode = True
    return retcode
    
    
    
def lte_mbim_is_connected(modem_drv):
    modem_drv = mbim_dev_check(modem_drv)
    retcode = False
    out = ""
    try:
        out = subprocess.run(['/usr/bin/mbim-network', modem_drv, 'status'], check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE).stdout.decode()
    except:
        return False
    m = re.search(".*Status: activated.*", out)
    if m:
        retcode = True
    return retcode



def save_lte_qmi_pdh(modem, apn, user, password, qmi_output):
    pdh = ""
    cid = ""
    pdh_match = re.search('.*handle:\s*(.*)\s*', str(qmi_output))
    cid_match = re.search('.*CID:\s*\'(.*)\'\s*', str(qmi_output))
    if pdh_match is not None and pdh_match.group(1) is not None:
        pdh = pdh_match.group(1)
        pdh = pdh.strip('\'"')

    if cid_match is not None and cid_match.group(1) is not None:
        cid = cid_match.group(1)

    lte_inf = ConfigObj()
    lte_inf['pdh'] = pdh
    lte_inf['cid'] = cid
    lte_inf['apn'] = apn
    lte_inf['user'] = user
    lte_inf['password'] = password
    lte_inf.filename = "/run/" + os.path.basename(modem) + ".prof"
    lte_inf.write()


def lte_putcmd(modem, cmd):
    print('Modem command: %s' % cmd, file=lteStream)
    modem.write((cmd + '\r').encode())
    time.sleep(0.5)


def lte_getresponse(modem, maxlines=3, timeout=1):
    buf = ''
    while True:
        modem.timeout = timeout
        l = modem.readline()
        if not l:
            break

        buf += l
        maxlines -= 1
        if maxlines <= 1:
            break

    print('Modem response %s' % buf, file=lteStream)
    return buf


def reset_modem_profile(tty_dev):
    try:
        modem_tty = serial.Serial(tty_dev, 115200, timeout=3)
        lte_putcmd(modem_tty, "AT+CGDCONT=1,\"IP\"")
        time.sleep(1)
        lte_getresponse(modem_tty)
        modem_tty.close()
        time.sleep(3)
    except:
        pass


def reset_lte_qmi_dev(modem):
    if not os.path.exists(modem):
        print('No device %s found' % modem, file=lteStream)
        return
    try:
        subprocess.check_output(['/usr/bin/qmicli', '-d', modem, '--dms-set-operating-mode=offline'])
        time.sleep(5)
        subprocess.check_output(['/usr/bin/qmicli', '-d', modem, '--dms-set-operating-mode=reset'])
        time.sleep(30)
    except:
        pass

def setup_lte_qmi_dev(if_cfg, modem):
    if not os.path.exists(modem):
        print('No device %s found' % modem, file=lteStream)
        return
        
    #if if_cfg['name'] != "lte0":
    #    return
    if lte_get_driver_type(if_cfg['name']) != QMI_DRV:
        return
    
    os.system("/usr/bin/qmicli -d %s --uim-sim-power-on=1" %modem)
    
    #store list of networks
    if not os.path.exists("/run/ltescan.txt"):
        os.system("/usr/bin/timeout 60s /usr/bin/qmicli -d %s --nas-network-scan > /run/ltescan.txt" %modem)
    
    ip_type = ""
    auth = ""
    apn, user, password = get_qmi_authenticators(if_cfg, modem)
    if os.path.exists(modem_tty_default):
        reset_modem_profile(modem_tty_default)

        
    start_cmd = ['/usr/bin/qmicli', '-d', modem,
                 '--wds-start-network=%s,%s,%s,%s,%s' % (apn, auth, user, password, ip_type),
                 '--client-no-release-cid', '--verbose', '--device-open-sync']
                 
    if user!="" and password!="":
        custom_cmd='/usr/bin/qmicli -d %s --wds-start-network="apn=%s,ip-type=4,username=%s,password=%s" --client-no-release-cid --verbose --device-open-sync' % (modem,apn, user, password)
        retcode = os.system(custom_cmd)
        if retcode == 0:
            print("Connected with cmd %s." % custom_cmd, file=lteStream)
            return True
            
    
    if user!="" and password!="" and apn!="":
        #try to set profile
        os.system('qmicli -d %s --wds-swi-create-profile-indexed="3gpp,1,apn=\'%s\',pdp-type=\'IP\',username=\'%s\',password=\'%s\'"' % (modem,apn, user, password) )
        os.system('qmicli -d %s --wds-set-default-profile-number=3gpp,1' % modem )
    else:
        os.system('qmicli -d %s --wds-delete-profile=3gpp,1' % modem )
    
    try:
        print("attempt to init LTE with apn '%s'" % apn, file=lteStream)
        out = subprocess.check_output(start_cmd, stderr=subprocess.STDOUT).decode()
        print(out, file=lteStream)
        save_lte_qmi_pdh(modem, apn, user, password, out)
        return True
    except subprocess.CalledProcessError as e:
        print(e.output, file=lteStream)
        start_cmd = ['/usr/bin/qmicli', '-d', modem,
                 '--wds-start-network="%s,%s,%s,%s,%s"' % (apn, auth, user, password, ip_type),
                 '--client-no-release-cid', '--verbose', '--device-open-sync']
        if "translated = [ type = '3gpp' reason = '50' ]" in str(e.output) or "ipv4-only-allowed" in str(e.output):
            start_cmd[3] = '--wds-start-network=\"'
            if apn != "":
                start_cmd[3] += "apn='%s'," % apn
            if user != "":
                start_cmd[3] += "username='%s'," % user
            if password != "":
                start_cmd[3] += "password='%s'," % password
            start_cmd[3] += 'ip-type=4\"'
            print(
                "Failed - operator doesn't support ipv6. attempt to init LTE %s with apn '%s', authentication '%s', user '%s', password '%s', and ip type '%s'" %
                (modem, auth, apn, user, password, ip_type), file=lteStream)
            retcode = os.system('/usr/bin/qmicli -d %s %s --client-no-release-cid --verbose --device-open-sync 1>/dev/null 2>/dev/null' % 
                                 (modem, start_cmd[3]) )
            if retcode == 0:
                print("Connected.", file=lteStream)
                return True
        print("Failed to connect %s using apn '%s', authentication '%s', user '%s', password '%s', and ip type '%s'" % (
            modem, apn, auth, user, password, ip_type), file=lteStream)

        try:
            out = subprocess.check_output(
                ['/usr/bin/qmicli', '-d', modem, '--wds-start-network=', '--client-no-release-cid', '--verbose',
                 '--device-open-sync'], stderr=subprocess.STDOUT).decode()
            save_lte_qmi_pdh(modem, apn, user, password, out)
            print("Connected.", file=lteStream)
            return True
        except subprocess.CalledProcessError:
            print('Failed to connect without APN, user and password', file=lteStream)
            return False

        return False
    except OSError as e:
        print('Error -- %s' % repr(e), file=lteStream)
        return False


def mbim_dev_check(modem=mbim_modem_default):
    if os.path.exists(modem):
        return modem
    if os.path.exists(modem_default):
        return modem_default


def setup_lte_mbim_dev(if_cfg, modem=mbim_modem_default):
    modem = mbim_dev_check(modem)
    retry = False
    if not os.path.exists(modem):
        print('No device %s found' % modem, file=lteStream)
        return
        
    drv_type = lte_get_driver_type(if_cfg['name'])
    if drv_type != MHI_DRV and drv_type != MBM_DRV:
        return
    
    apn = if_cfg['apn'] if 'apn' in if_cfg else ""
    user = if_cfg['user'] if 'user' in if_cfg else ""
    password = if_cfg['pass'] if 'pass' in if_cfg else ""

    start_cmd = ['/usr/bin/mbim-network',  modem, 'start']
    
    with open('/etc/mbim-network.conf', 'w') as f:
        if apn != "":
             f.write("APN=%s\n" % apn)
        else:
            f.write("APN=internet\n" )
        if user !="" and password !="":
             f.write("APN_USER=%s APN_PASS=%s\n" %(user,password))
             f.write("APN_AUTH=pap\n")
        f.write("PROXY=yes\n")
    
    try:
        print("attempt to init LTE ", file=lteStream)
        out = subprocess.check_output(start_cmd, stderr=subprocess.STDOUT).decode()
        print(out, file=lteStream)
        return True
    except subprocess.CalledProcessError as e:
        print(e.output, file=lteStream)
        retry = True
        return False
    
    if retry:
        time.sleep(3)
        try:
            out = subprocess.check_output(start_cmd, stderr=subprocess.STDOUT).decode()
            return True
        except subprocess.CalledProcessError as e:
            return False
    
    return False
    

def disconnect_lte_qmi_dev(modem, pdh):
    if pdh is None or pdh == "":
        return
    stop_cmd = ['/usr/bin/qmicli', '-d', modem, '--wds-stop-network=%s' % pdh]
    for _ in range(3):
        try:
            subprocess.check_output(stop_cmd, stderr=subprocess.STDOUT).decode()
        except subprocess.CalledProcessError:
            time.sleep(2)
        if not lte_qmi_is_connected(modem):
            break

    try:
        subprocess.check_output(stop_cmd, stderr=subprocess.STDOUT).decode()
    except:
        pass

    time.sleep(5)


def load_lte_qmi_pdh(modem):
    apn = ""
    user = ""
    password = ""
    pdh = ""
    cid = ""
    lte_inf = ConfigObj("/run/" + os.path.basename(modem) + ".prof")
    if 'pdh' in lte_inf:
        pdh = lte_inf['pdh']
    if 'cid' in lte_inf:
        cid = lte_inf['cid']
    if 'apn' in lte_inf:
        apn = lte_inf['apn']
    if 'user' in lte_inf:
        user = lte_inf['user']
    if 'password' in lte_inf:
        password = lte_inf['password']
    return apn, user, password, pdh, cid


def reconnect_lte_qmi_dev(if_cfg, modem, force=False):
    if not lte_is_qmi(modem):
        return

    if not os.path.exists(modem):
        return

    apn, user, password = get_qmi_authenticators(if_cfg, modem)
    (oldapn, olduser, oldpassword, pdh, cid) = load_lte_qmi_pdh(modem)

    if force is False:
        if not lte_qmi_is_connected(modem):
            return

        if apn == oldapn and user == olduser and password == oldpassword:
            return

    if pdh is None or pdh == "":
        return

    disconnect_lte_qmi_dev(modem, pdh)

    setup_lte_qmi_dev(if_cfg, modem)
    
    
    
def lte_check_raw_if(if_name):
    if not os.path.exists('/sys/class/net/' + if_name + '/qmi/raw_ip'):
        if os.path.exists('/sys/class/net/' + if_name + '/type'):
            raw_type = open('/sys/class/net/' + if_name + '/type', 'rb').readline().rstrip().decode()
            if raw_type == '519':
                return True
        return False

    raw_mode = open('/sys/class/net/' + if_name + '/qmi/raw_ip', 'rb').readline().rstrip().decode()
    if raw_mode == 'Y':
        return True

    return False



def lte_get_raw_mac(ifname):
    fake_mac = '00:00:00:00:a1:00'

    def extract_mac_from_imei(imei_line):
        try:
            imei = imei_line.split(':')
            imei_int = int(imei[-1].replace(' ', '').replace('\'', ''))
            (_, mac) = divmod(imei_int, 0x1000000000000)
            mac_str = '{:012x}'.format(mac)
            octet1 = int(mac_str[0:2], 16) & 0xfe
            octet1_str = "%02X" % octet1
            return '{}:{}:{}:{}:{}:{}'.format(octet1_str, mac_str[2:4], mac_str[4:6], mac_str[6:8], mac_str[8:10],
                                                  mac_str[10:12])
        except:
            return fake_mac
            
    if not os.path.exists('/sys/class/net/' + ifname + '/qmi'):
        if os.path.exists('/dev/cdc-wdm_pcie'):
            try:
                out = subprocess.check_output(['mbimcli', '--device=/dev/cdc-wdm_pcie', '--device-open-proxy' , '--query-device-caps']).decode()
                for line in out.split('\n'):
                    if 'Device ID' in line:
                        return extract_mac_from_imei(line)
            except:
                pass
            return fake_mac
        return fake_mac

    try:
        out = subprocess.check_output(['qmicli', '-d', '/dev/cdc-wdm' + ifname.strip('lte'), '--dms-get-ids']).decode()
        for line in out.split('\n'):
            if 'IMEI' in line:
                return extract_mac_from_imei(line)
    except:
        return fake_mac

#Get type of the modem driver per interface name
#Input: name of the interface , i.e.: lte0
#Output: 1 - QMI ; 2 - MBIM ; 3 - CDC-Ether (dongle) ; 10 - Unknown

def lte_get_driver_type(ifname):
    retcode = 10
    out = ""
    try:
        out = subprocess.run(['/sbin/ethtool', '-i', ifname], check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE).stdout.decode()
    except:
        return retcode
    if re.search(".*driver: mhi.*", out):
        retcode=MHI_DRV
    elif re.search(".*driver: qmi_wwan.*", out):
        retcode=QMI_DRV
    elif re.search(".*driver: cdc_ether.*", out):
        retcode=CDC_DRV
    elif re.search(".*driver: cdc_mbim.*", out):
        retcode=MBM_DRV
    
    return retcode



def lte_get_driver_type_by_dev(devpath):
    retcode = 10
    out = ""
    devname=devpath.replace("/dev/","")
    
    
    uevent_name='/sys/class/usbmisc/' + devname + '/device/uevent'
    if not os.path.exists(uevent_name):
        uevent_name='/sys/class/mhi_uci/' + devname + '/device/uevent'
        if not os.path.exists(uevent_name):
            return retcode
    
    
    try:
        out = subprocess.run(["cat", uevent_name], check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE).stdout.decode()
    except:
        return retcode
    if re.search(".*DRIVER=mhi.*", out):
        retcode=MHI_DRV
    elif re.search(".*DRIVER=qmi_wwan.*", out):
        retcode=QMI_DRV
    elif re.search(".*DRIVER=cdc_ether.*", out):
        retcode=CDC_DRV
    elif re.search(".*DRIVER=cdc_mbim.*", out):
        retcode=MBM_DRV
    return retcode
