#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import sys
import subprocess

from netaddr import *
from systemd import journal
from tveif import if_eth

RTINDEX_OFFSET = 1
ifNatStream = journal.stream(identifier='if_nat')
sysrepo_file = "/config/_sysrepo"


class if_nat(if_eth.if_eth):
    def __init__(self, if_cfg=None):
        self.nat_set = False
        self.name = None
        self.ipv4addr = None
        self.ipv6addr = None
        super(if_nat, self).__init__(self.name, if_cfg)
        self.wan = []

        if if_cfg is not None:
            # Getting the wan interfaces
            for wan_if in if_cfg.get("wan-interface", []):
                if 'data' in wan_if.get('name', None):
                    self.wan.append('br-%s' % wan_if.get("name", None))
                else:
                    self.wan.append(wan_if.get("name", None))
            
            # Getting the IP address from nat cfg
            if os.path.exists(sysrepo_file):
                ipv4addr = if_cfg.get('ipv4-prefix', '')
                ipv6addr = if_cfg.get('ipv6-prefix', '')
                if '' == ipv4addr and '' == ipv6addr:
                    self.bootproto = 'manual'
                else:
                    self.bootproto = 'static'
                    if ipv4addr: 
                        self.ipv4addr = str(IPNetwork(ipv4addr).ip)
                    if ipv6addr:
                        self.ipv6addr = str(IPNetwork(ipv6addr).ip)
            else:
                ipv4addr = if_cfg.get('ipv4-address', '')
                ipv6addr = if_cfg.get('ipv6-address', '')
                if '' == ipv4addr and '' == ipv6addr:
                    self.bootproto = 'manual'
                else:
                    self.bootproto = 'static'
                    if ipv4addr: 
                        self.ipv4addr = ipv4addr
                    if ipv6addr: 
                        self.ipv6addr = ipv6addr

    def get_name(self):
        if self.name is None:
            return "NONE"
        else:
            return self.name

    def get_sys_default(self):
        default_rt = []
        # getting default routes for IPv4
        iproutes = subprocess.check_output(['/sbin/ip', 'route', 'list', 'exact', '0.0.0.0/0']).decode()
        # getting default routes for IPv6
        iproutes6 = subprocess.check_output(['/sbin/ip', '-6', 'route', 'list', 'exact', '::/0']).decode()

        for route in iproutes.splitlines():
            rt_line = route.split(' ')
            rt_line = [x for x in rt_line if x != '']
            # Example data for rt_line:
                # default via 192.168.8.1 dev lte0  metric 100
                # default via fe80::1111:2222 dev lte0 proto ra metric 1024 expires 10sec hoplimit 64 pref medium
            if rt_line[4] in self.wan:
                #                        address     br-data     metric
                default_rt.append(tuple([rt_line[2], rt_line[4], None if len(rt_line) < 7 else rt_line[6]]))
        for route6 in iproutes6.splitlines():
            rt_line = route6.split(' ')
            rt_line = [x for x in rt_line if x != '']
            if rt_line[4] in self.wan:
                #                        addr        br-data     metric
                default_rt.append(tuple([rt_line[2], rt_line[4], rt_line[8] if 'metric' in rt_line else None]))
                #                                                if metric is found in routing table line, give metric
        return default_rt

    @staticmethod
    def reset_nat_rt(if_name, nat, gw, metric):
        # used for resetting nat's routing table when wan-interface of nat has been configured after nat
        tableid = int(nat) + RTINDEX_OFFSET
        try:
            cmd = ['add', 'default', 'via', str(gw), 'dev', str(if_name)]
            # If metric is available, append it to ip route command
            if metric and metric != "none":
                cmd.append('metric')
                cmd.append(str(metric))
            cmd.append('table')
            cmd.append(str(tableid))
            
            # Adapt ip route command depending on IP version
            if gw.find(":") > -1:
                cmd = ['/sbin/ip', '-6', 'route'] + cmd
            else:
                cmd = ['/sbin/ip', 'route'] + cmd
                    
            subprocess.check_call(cmd)

        except subprocess.CalledProcessError as e:
            print('\"%s\" Failed with code %d - \"%s\"' % (e.cmd, e.returncode, e.output), file=ifNatStream)
            
        #if we have load balancing, now add ECMP route with metric 1  
        #this is demo only for NanoPI as a home router
        if os.path.isfile("/config/nat-lb"):
            os.system("ip route del  default metric 1  table %d" % tableid)
            cmd='ip route add default proto static metric 1 table %d $(for metric in `seq 2 249` ; do  ip route list default table %d metric $metric ; done | sed \'s/default/nexthop/\'  | sed \'s/proto static/weight 1/\' | tr "\n" " ")' %  (tableid, tableid)
            os.system(cmd)

    def set_nat_rt(self):
        tableid = int(self.name.replace('nat', '')) + RTINDEX_OFFSET

        # set rules in table tableid
        try:
            # IPv4 rules
            subprocess.check_call(['/sbin/ip', 'rule', 'add', 'iif', str(self.name), 'unreachable'])
            subprocess.check_call(['/sbin/ip', 'rule', 'add', 'iif', str(self.name), 'table', str(tableid)])
            # IPv6 rules
            subprocess.check_call(['/sbin/ip', '-6', 'rule', 'add', 'iif', str(self.name), 'unreachable'])
            subprocess.check_call(['/sbin/ip', '-6', 'rule', 'add', 'iif', str(self.name), 'table', str(tableid)])
        except subprocess.CalledProcessError as e:
            print('\'%s\' Failed with code %d - \"%s\"' % (e.cmd, e.returncode, e.output), file=ifNatStream)

        for (gw, ifname, metric) in self.get_sys_default():
            try:
                # Adapt ip route command depending on IP version
                if ":" in gw:
                    iproute_version = ['/sbin/ip', '-6', 'route']
                else:
                    iproute_version = ['/sbin/ip', 'route']
                
                cmd = iproute_version + ['add', 'default', 'via', str(gw), 'dev', str(ifname), 'table', str(tableid)]

                # Append metric to ip route command if it's available
                if metric:
                    cmd.insert(len(cmd)-2, 'metric')
                    cmd.insert(len(cmd)-2, str(metric))

                subprocess.check_call(cmd)

            except subprocess.CalledProcessError as e:
                print('\"%s\" Failed with code %d - \"%s\"' % (e.cmd, e.returncode, e.output), file=ifNatStream)
            except:
                print('Error occured while setting routing table for NAT.', file=ifNatStream)

    def set_nat(self):
        if not self.wan:
            return

        for wan in self.wan:
            try:
                subprocess.check_call(
                    ['/sbin/iptables', '-t', 'nat', '-A', 'POSTROUTING', '-o', wan, '-j', 'MASQUERADE'])
                subprocess.check_call(
                    ['/sbin/iptables', '-A', 'FORWARD', '-i', wan, '-o', str(self.name), '-m', 'state', '--state',
                     'RELATED,ESTABLISHED', '-j', 'ACCEPT'])
                subprocess.check_call(
                    ['/sbin/iptables', '-A', 'FORWARD', '-i', str(self.name), '-o', wan, '-j', 'ACCEPT'])
            except subprocess.CalledProcessError as e:
                print('set_nat -> \"%s\" Failed with code %d - \"%s\"' % (e.cmd, e.returncode, e.output),
                      file=ifNatStream)

        self.set_nat_rt()

    def clean_nat_rt(self):
        tableid = int(self.name.replace('nat', '')) + RTINDEX_OFFSET
        try:
            # IPv4
            subprocess.check_call(['/sbin/ip', 'rule', 'del', 'iif', str(self.name), 'table', str(tableid)])
            subprocess.check_call(['/sbin/ip', 'rule', 'del', 'iif', str(self.name), 'unreachable'])
            # IPv6
            subprocess.check_call(['/sbin/ip', '-6', 'rule', 'del', 'iif', str(self.name), 'table', str(tableid)])
            subprocess.check_call(['/sbin/ip', '-6', 'rule', 'del', 'iif', str(self.name), 'unreachable'])
        except subprocess.CalledProcessError as e:
            print('\"%s\" Failed with code %d - \"%s\"' % (e.cmd, e.returncode, e.output), file=ifNatStream)

        try:
            # IPv4
            subprocess.check_call(['/sbin/ip', 'route', 'flush', 'table', str(tableid)])
            # IPv6
            subprocess.check_call(['/sbin/ip', '-6', 'route', 'flush', 'table', str(tableid)])
        except subprocess.CalledProcessError as e:
            print('\"%s\" Failed with code %d - \"%s\"' % (e.cmd, e.returncode, e.output), file=ifNatStream)

    def clean_nat(self):
        if not self.wan:
            return

        for wan in self.wan:
            try:
                print('clean_nat -D FORWARD 1 for %s' % wan, file=ifNatStream)
                subprocess.check_output(
                    ['/sbin/iptables', '-D', 'FORWARD', '-i', wan, '-o', self.name, '-m', 'state', '--state',
                     'RELATED,ESTABLISHED', '-j', 'ACCEPT'], stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                print('\"%s\" Failed with code %d - \"%s\"' % (e.cmd, e.returncode, e.output), file=ifNatStream)

            try:
                print('clean_nat -D FORWARD 2 for %s' % wan, file=ifNatStream)
                subprocess.check_output(['/sbin/iptables', '-D', 'FORWARD', '-i', self.name, '-o', wan, '-j', 'ACCEPT'],
                                        stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                print('\"%s\" Failed with code %d - \"%s\"' % (e.cmd, e.returncode, e.output), file=ifNatStream)

            try:
                print('clean_nat -t nat for %s' % wan, file=ifNatStream)
                subprocess.check_output(
                    ['/sbin/iptables', '-t', 'nat', '-D', 'POSTROUTING', '-o', wan, '-j', 'MASQUERADE'],
                    stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                print('\"%s\" Failed with code %d - \"%s\"' % (e.cmd, e.returncode, e.output), file=ifNatStream)

        self.clean_nat_rt()

    def get_wan(self):
        return list(self.wan)

    def set_wan(self, list_wan):
        re_init = False
        self.wan = list(list_wan)

    def create(self, dpdk=False):
        self.dpdk_mode = dpdk

        try:
            subprocess.check_call(['/usr/bin/ovs-vsctl', 'add-br', self.name])
        except subprocess.CalledProcessError as e:
            print('Command \"%s\" failed with code %d' % (e.cmd, e.returncode), file=ifNatStream)

        self.write_ifcfg()

    def write_ifcfg(self, fname=None):
        createBridgeIfString = [
            "/scripts/apply_networking.py",
            "-n%s" % self.name,
            "-b%s" % self.bootproto,
            "--ovs-master", "ovs"]
        if None != self.ipv4addr:
            createBridgeIfString += ["-a%s" % self.ipv4addr]
        if None != self.ipv6addr:
            createBridgeIfString += ["-f%s" % self.ipv6addr]

        try:
            subprocess.check_call(createBridgeIfString)
        except subprocess.CalledProcessError as e:
            print("Command \"%s\" failed with code %d" % (e.cmd, e.returncode), file=ifNatStream)
        except OSError as e:
            print('System error - \"%s\"' % e.strerror, file=ifNatStream)

    def delete(self):
        subprocess.check_call(['/sbin/ifdown', self.name])
        if self.cfg_fname is not None:
            os.remove(self.cfg_fname)

        self.clean_nat()
        subprocess.check_call(['/usr/bin/ovs-vsctl', 'del-br', self.name])


if __name__ == "__main__":
    # Used in dhclient-event-process.sh
    if len(sys.argv)>=5:
        if_name = sys.argv[1]
        nat_id = sys.argv[2]
        gw = sys.argv[3]
        metric = sys.argv[4]
        if_nat.reset_nat_rt(if_name, nat_id, gw, metric)
